from functools import cache
import sympy as sym
from .base_element import BaseElement

class Damper(BaseElement):
    def __init__(self,q,velocity,damping_constant,name = "default"):
        self.__c = damping_constant
        self.__z_dot = velocity
        super(Damper, self).__init__(q,name)
    
    ke = property(lambda self: 0)
    pe = property(lambda self: 0)
    M = property(lambda self: sym.zeros(len(self.q)))

    @property
    @cache
    def rdf(self):
        return sym.Rational(1,2)*self.__c*self.__z_dot**2


from functools import cache
from moyra.elements.mass_matrix import MassMatrix
from moyra.frames.homogenous_frame import HomogenousFrame
from moyra.frames.reference_frame import ReferenceFrame
from moyra.helper_funcs import Wedge
import sympy as sym
from sympy.abc import t
from .base_element import BaseElement
from sympy.physics.mechanics import msubs,dot
from collections.abc import Iterable

class ModalElement(BaseElement):
    def __init__(self,q,frame,q_f,M_trace,K_trace,simplify=True, name="default"):

        self._frame = frame
        self._q_f = sym.Matrix(q_f) if isinstance(q_f,Iterable) else sym.Matrix([q_f])
        self._M_trace = M_trace
        self._K_trace = K_trace

        if isinstance(frame,HomogenousFrame):
            q = q
        else:
            raise ValueError('Invlaid Transform type')

        super(ModalElement, self).__init__(q,name)
        self._simplify = simplify
    
    q_f = property(lambda self: self._q_f)
    frame = property(lambda self: self._frame)
    M_trace = property(lambda self: self._M_trace)
    K_trace = property(lambda self: self._K_trace)
    simplify = property(lambda self: self._simplify)

    @property
    @cache
    def ke(self):
        # calculate the K.E
        T = sym.Rational(1,2)*self.qd.T*self.M*self.qd
        return T[0]
    
    @property
    @cache
    def pe(self):
        return self.elastic_pe + self.grav_pe

    @property
    @cache
    def elastic_pe(self):
        U = sym.Rational(1,2)*self.q_f.T*sym.diag(*self.K_trace)*self.q_f
        return U[0]
    
    @property
    @cache
    def grav_pe(self):
        return sym.Integer(0)

    @property
    def rdf(self):
        return sym.Integer(0)

    @property
    @cache
    def M(self):
        M = sym.zeros(len(self.q))
        idx = [i for i,val in enumerate(self.q_f) if val in self.q]
        for i in idx:
            M[i,i] = self.M_trace[i]
        return M

    def _trigsimp(self,expr):
        return sym.trigsimp(sym.powsimp(sym.cancel(sym.expand(expr))))



            



import sympy as sym
from . import ExternalForce

class GravityForce(ExternalForce):

    def __init__(self,q,Transform,ForceVector):
        # create the wrench applied at the origin of the endeffector in spatial coords
        wrench_g = sym.Matrix([ForceVector[0],ForceVector[1],ForceVector[2],0,0,0])

        ## convert this to a spatial wrench

        # make a frame a end effector in body frame
        T_trans = Transform.PuesdoSpatialFrame()

        # convert wrench into this frame
        F_s = T_trans.Adjoint().T*wrench_g

        # convert into joint torques
        _Q = sym.simplify(T_trans.ManipJacobian(q).T*F_s)

        super().__init__(_Q)



from moyra.frames.base_reference_frame import BaseReferenceFrame
import sympy as sym
from sympy.abc import t
import sympy.physics.mechanics as me

from moyra.helper_funcs import Wedge


class ReferenceFrame(BaseReferenceFrame):

    @classmethod
    def EulerParameters(cls,R,theta):
        G = sym.Integer(2)*sym.Matrix([[-theta[1],theta[0],-theta[3],theta[2],],
                                        [-theta[2],theta[3],theta[0],-theta[1],],
                                        [-theta[3],-theta[2],theta[1],theta[0],]])
        Gb = sym.Integer(2)*sym.Matrix([[-theta[1],theta[0],theta[3],-theta[2],],
                                        [-theta[2],-theta[3],theta[0],theta[1],],
                                        [-theta[3],theta[2],-theta[1],theta[0],]])
        A = (sym.Rational(1,2)*G)*(sym.Rational(1,2)*Gb).T                                        
        return cls(theta,A,R,G,Gb)

    @classmethod
    def EulerAnglesXYZ(cls,R,theta):
        G = sym.Matrix([  [1,0,sym.sin(theta[1])],
                                [0,sym.cos(theta[0]),-sym.sin(theta[0])*sym.cos(theta[1])],
                                [0,sym.sin(theta[1]),sym.cos(theta[0])*sym.cos(theta[1])]])
        Gb = sym.Matrix([  [sym.cos(theta[1])*sym.cos(theta[2]),sym.sin(theta[2]),0],
                                [-sym.sin(theta[2])*sym.cos(theta[1]),sym.cos(theta[2]),0],
                                [sym.sin(theta[1]),0,1]])
        A = sym.rot_axis1(-theta[0])*sym.rot_axis2(-theta[1])*sym.rot_axis3(-theta[2])
        return cls(theta,A,R,G,Gb)

    def __init__(self,theta,A,R,G,Gb):
        self._theta = sym.Matrix([*theta])
        self._G = G.copy()
        self._Gb = Gb.copy()
        super(ReferenceFrame,self).__init__(A,R)
    
    G = property(lambda self: self._G)
    Gb = property(lambda self: self._Gb)
    theta = property(lambda self: self._theta)

    def GlobalVelocity(self,u=sym.Matrix([0]*3)):
        return self.R.diff(t)-self.A*Wedge(u)*self.Gb*self.theta.diff(t)

    def BodyVelocity(self,u=sym.Matrix([0]*3)):
        return self.A.T*self.GlobalVelocity(u)

    def BodyJacobian(self,q,u=sym.Matrix([0]*3)):
        B = sym.zeros(3,len(self.theta))
        for i in range(len(self.theta)):
            B[:,i] = (self.A*u).diff(self.theta[i])
        Jb = [[self.A.T,self.A.T*B],[sym.zeros(3),self.Gb]]
        return sym.simplify(sym.BlockMatrix(Jb).as_explicit())
    
    def PuesdoSpatialFrame(self):
        return ReferenceFrame( self.theta.copy(),sym.eye(3),
                                self.t.copy(),sym.zeros(3,len(self.theta)),
                                sym.zeros(3,len(self.theta)))

    def simplify(self):
        return ReferenceFrame(self.theta.copy(),sym.simplify(self.R),
                                sym.simplify(self.t),
                                sym.simplify(self.G),
                                sym.simplify(self.Gb))

    def subs(self,*args):  
        return ReferenceFrame(self.theta.subs(*args),self.R.subs(*args),
                                self.t.subs(*args),self.G.subs(*args),
                                self.Gb.subs(*args))

    def msubs(self,*args):  
        return ReferenceFrame(me.msubs(self.theta,*args),me.msubs(self.R,*args),
                                me.msubs(self.t,*args),me.msubs(self.G,*args),
                                me.msubs(self.Gb,*args))


from functools import cache
from moyra.elements.mass_matrix import MassMatrix
from moyra.frames.homogenous_frame import HomogenousFrame
from moyra.frames.reference_frame import ReferenceFrame
from moyra.helper_funcs import Wedge
import sympy as sym
from sympy.abc import t
from .base_element import BaseElement
from sympy.physics.mechanics import msubs,dot
from collections.abc import Iterable

class FlexiElement(BaseElement):
    def __init__(self,q,frame,density,S,x_integral,y_integral,x_f,q_f,EI,GJ,grav_vec=sym.Matrix([0]*3) ,simplify=True, name="default"):
        self._EI = EI
        self._GJ = GJ
        self._x_f = x_f
        self._x_integral = x_integral
        self._y_integral = y_integral

        self._q_f = sym.Matrix(q_f) if isinstance(q_f,Iterable) else sym.Matrix([q_f])
        self._S = S
        self._frame = frame

        if isinstance(frame,HomogenousFrame):
            q = q
        elif isinstance(frame,ReferenceFrame):
            q = sym.Matrix([*frame.R,*frame.theta,*q_f])
        else:
            raise ValueError('Invlaid Transform type')

        super(FlexiElement, self).__init__(q,name)

        self._grav_vec = sym.Matrix(grav_vec)
        self._density = density
        self._simplify = simplify
    
    x = property(lambda self: self.x_integral[0])
    y = property(lambda self: self.y_integral[0])
    EI = property(lambda self: self._EI)
    GJ = property(lambda self: self._GJ)
    x_f = property(lambda self: self._x_f)
    x_integral = property(lambda self: self._x_integral)
    y_integral = property(lambda self: self._y_integral)
    q_f = property(lambda self: self._q_f)
    S = property(lambda self: self._S)
    frame = property(lambda self: self._frame)
    grav_vec = property(lambda self: self._grav_vec)
    density = property(lambda self: self._density)
    simplify = property(lambda self: self._simplify)

    @property
    @cache
    def u_0(self):
        return sym.Matrix([self.x,self.y,0])
        
    @property
    @cache
    def u_f(self):
        return self.S*self.q_f

    @property
    @cache
    def ke(self):
        if self.density==0:
            return 0
        # calculate the K.E
        T = sym.Rational(1,2)*self.qd.T*self.M*self.qd
        return T[0].expand()
    
    @property
    @cache
    def pe(self):
        return self.elastic_pe + self.grav_pe

    @property
    @cache
    def elastic_pe(self):
        # Bending Potential Energy per unit length
        U_e = 0
        if isinstance(self.EI, sym.Expr) or self.EI != 0:
            v = msubs(self.u_f,{self.x:self.x_f}).diff(self.y,self.y)
            U_e += self._trigsimp((v.T*v))[0]*self.EI*sym.Rational(1,2)

        # Torsional P.E per unit length
        if isinstance(self.GJ, sym.Expr) or self.GJ != 0:
            v = msubs(self.u_f.diff(self.x),{self.x:self.x_f}).diff(self.y)
            U_e += self._trigsimp((v.T*v))[0]*self.GJ*sym.Rational(1,2)

        return U_e.integrate(self.y_integral) if isinstance(U_e, sym.Expr) else U_e
    
    @property
    @cache
    def grav_pe(self):
        if (self.grav_vec.T*self.grav_vec)[0] == 0:
            return 0
        elif isinstance(self.frame,HomogenousFrame):   
            point = self.frame.R
            dmg = -(point.T*self.grav_vec)[0]*self.density
            return dmg.integrate(self.x_integral,self.y_integral) if dmg != 0 else 0
        elif isinstance(self.frame,ReferenceFrame):  
            g = -self.grav_vec
            mg = self.density.integrate(self.x_integral,self.y_integral)*self.frame.R.T*g
            mg += (self.density*(self.u_0 + self.u_f).T).integrate(self.x_integral,self.y_integral)*self.frame.A.T*g
            return mg[0]
        else:
            raise ValueError('Transform must either be type HomogenousFrame or ReferenceFrame')


    @property
    def rdf(self):
        return sym.Integer(0)

    @property
    @cache
    def M(self):
        # create the jacobian for the mass 
        if isinstance(self.frame,HomogenousFrame):   
            Js = self.frame.ManipJacobian(self.q)
            Jb = self.frame.InvAdjoint()*Js
            Jb = self._trigsimp(Jb)
            M_e = MassMatrix(self.density)
            M = (Jb.T*M_e*Jb).integrate(self.x_integral,self.y_integral)
        elif isinstance(self.frame,ReferenceFrame):
            M_rr = sym.eye(3)*(self.density.integrate(self.x_integral,self.y_integral))
            u = Wedge(self.u_0 + self.u_f)
            A = self.frame.A
            Gb = self.frame.Gb
            S = self.S

            M_rt = -A
            M_rt *= (self.density*u).integrate(self.x_integral,self.y_integral)
            M_rt *= Gb

            M_rf = A * (self.density * S).integrate(self.x_integral,self.y_integral)
            M_tt = Gb.T
            M_tt *= (self.density*u.T*u).integrate(self.x_integral,self.y_integral)
            M_tt *= Gb

            M_tf = Gb.T
            M_tf *= (self.density*u*S).integrate(self.x_integral,self.y_integral)

            M_ff = (self.density*S.T*S).integrate(self.x_integral,self.y_integral)

            M = sym.BlockMatrix([[M_rr,M_rt,M_rf],[M_rt.T,M_tt,M_tf],[M_rf.T,M_tf.T,M_ff]]).as_explicit()
        else:
            raise ValueError("Transform must either be type HomogenousFrame or ReferenceFrame")
        return self._trigsimp(M) if self.simplify else M

    def _trigsimp(self,expr):
        return sym.trigsimp(sym.powsimp(sym.cancel(sym.expand(expr))))

    @staticmethod
    def ShapeFunctions_OBM_IBN_TO(m,n,o,q,y_s,x_s,x_f,factor = 1,type='taylor'):
        """
        returns the shape function mAtrix for an assumed shapes beam
        
        The function returns a tple of the shape function matrix S and the twist
         of an assumed shapes beam with 'm' out-of-plane bending shpaes, 'n' 
         in-plane bending shapes, and 'o' twist shapes, the list 'q' represents the DoF in the order n,m,o 
        
        Parameters
        ----------
        m : int
            number of out-of-plane bending shapes
        n : int
            number of in-plane bending shapes
        o : int 
        number of twist shapes
        q : list
            list of DoFs
        y_s : Symbol
            the y direction of the beam (the beam is integrated along its y direction)
        x_s : Symbol
            the x direction of the beam (XY plane is in plane bending)
        x_f : Symbol
            the location of the flexural axis e.g twist = 0 @ x_s=x_f
        factor : int / float / list
            scaling factor to apply to each DoF. If it is a single number
            it is applied to all DoFs, otherwise it must be a list equal in
            length to q
        type : str
            either 'taylor' or 'cheb' and defines the shape functions used 
            (talyor series or chebeshev series)

        Returns
        -------
        S:Dense Matrix
            A 3xlen(q) Matrix which defines the shape functions. Sq = v where v 
            is a 3x1 vector defining the deformed shape at a given undeformed 
            x_s and y_s position
        tau: Sympy Function
            a function defining the twist along the flexural axis
         """
        # ensure q is iterable even if only one element
        q = q if isinstance(q,Iterable) else [q]

        # check elements in q the same as n+m+o
        qs = len(q)
        if n+m+o != qs:
            raise ValueError('the sum of n+m+o must be the same as a length of q')
        S = sym.zeros(3,qs)
        # make factor a list the size of n+m
        if isinstance(factor,int) | isinstance(factor,float):
            factor = [factor]*(qs)

        # Out of plane bending
        for i in range(0,m):
            if type == 'taylor':
                S[2,i] += y_s**(2+i)*factor[i]
            elif type == 'cheb':
                S[2,i] += sym.chebyshevt_poly(i,y_s)*factor[i]
            else:
                raise ValueError('poly type must be either cheb or taylor')
        # In-plane bending
        for i in range(0,n):
            qi = m+i
            if type == 'taylor':
                S[0,qi] += y_s**(2+i)*factor[qi]
            elif type == 'cheb':
                S[0,qi] += sym.chebyshevt_poly(i,y_s)*factor[qi]
            else:
                raise ValueError('poly type must be either cheb or taylor')
        # Twist
        tau = 0
        for i in range(0,o):
            qi = i+m+n
            if type == 'taylor':
                S[2,qi] += -y_s**(i+1)*factor[qi]*(x_s-x_f)
                tau += q[qi]*y_s**(i+1)*factor[qi]
            elif type == 'cheb':
                S[2,qi] += -sym.chebyshevt_poly(i,y_s)*factor[qi]*(x_s-x_f)
                tau += q[qi]*sym.chebyshevt_poly(i,y_s)*factor[qi]

        return sym.simplify(S), sym.simplify(tau)




            



import sympy as sym
from functools import cache
from .base_element import BaseElement

class Spring(BaseElement):
    def __init__(self,q,deflection,spring_constant,name="default"):
        self.__k = spring_constant
        self.__z = deflection
        super(Spring, self).__init__(q,name)
        
    ke = property(lambda self:0)
    rdf = property(lambda self:0)
    M = property(lambda self: sym.zeros(len(self.q)))

    @property
    @cache
    def pe(self):
        return sym.Rational(1,2)*self.__k*self.__z**2
        



import shutil
from os import path
from setuptools import find_packages, setup
import os 


dir_path = path.dirname(path.realpath(__file__))

with open(path.join(dir_path,'version.txt'),'r') as f:
    ver = f.read()

# clean up directories
dirs = ['.eggs','build','dist','moyra.egg-info']
for _dir in dirs:
    if path.exists(path.join(dir_path,_dir)):
        shutil.rmtree(path.join(dir_path,_dir))



setup(
    name='moyra',
    packages=find_packages(include=['moyra','moyra.*']),
    version=ver,
    description='Generate Multi-body Symbolic and Numeric Equations of Motion',
    long_description = open(path.join(dir_path,'README.md')).read(),
    long_description_content_type="text/markdown",
    author='Fintan Healy',
    author_email = 'fintan.healy@bristol.ac.uk',
    url='https://github.com/fh9g12/moyra',
    license='MIT',
    install_requires=['sympy<=1.12','numpy<=1.26','scipy'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
    test_suite='tests',
)




from .rigid_element import RigidElement
from .flexi_element import FlexiElement
from .mass_matrix import MassMatrix
from .spring import Spring
from .damper import Damper
from .modal_element import ModalElement


from functools import cache
from moyra.frames.homogenous_frame import HomogenousFrame
from moyra.frames.reference_frame import ReferenceFrame
from moyra.helper_funcs import Wedge,Vee
import sympy as sym
from sympy.abc import t
from .base_element import BaseElement
from .mass_matrix import MassMatrix

class RigidElement(BaseElement):

    @classmethod
    def point_mass(cls,q,frame,m,grav_vec=sym.Matrix([0]*3),simplify=True):
        return cls(q,frame,MassMatrix(m),grav_vec,simplify=True)  

    def __init__(self, q, frame, M, grav_vec=sym.Matrix([0]*3), com_pos=[0,0,0], simplify=True, name = "default", alt_method=False):
        self._grav_vec = grav_vec
        self._frame = frame
        self._M_e = M
        self._com_pos = com_pos
        self._simplify = simplify
        self._alt_method = alt_method
        if isinstance(frame,HomogenousFrame):
            q = q
            self._idx =  list(range(len(q)))
        elif isinstance(frame,ReferenceFrame):
            q = sym.Matrix([*frame.R,*frame.theta])
            self._idx = [i for i,v in enumerate(q) if (v.args[0] == t if isinstance(v,sym.Function) else False)]
            q = sym.Matrix(q[self._idx,:])
        else:
            raise ValueError('Invlaid Transform type')

        super(RigidElement, self).__init__(q,name)
    
    M_e = property(lambda self: self._M_e)
    com_pos = property(lambda self: self._com_pos)
    frame = property(lambda self: self._frame)
    grav_vec = property(lambda self: self._grav_vec)
    simplify = property(lambda self: self._simplify)
    rdf = property(lambda self: 0)
    alt_method = property(lambda self: self._alt_method)

    @property
    @cache
    def ke(self):
        if self.M_e[0,0]==0:
            return 0
        if isinstance(self.frame,HomogenousFrame):
            Rf = self.frame.Translate(*self.com_pos)
        if isinstance(self.frame,ReferenceFrame):
            Rf = self.frame
        W = sym.Matrix([*Rf.R.diff(t),*Vee(Rf.A.T*Rf.A.diff(t))])
        T =  (sym.Rational(1,2) * W.T * self.M_e * W)[0]
        return T.expand() if self.simplify else T

    @property
    @cache
    def M(self):
        #get M in world frame
        #calculate the mass Matrix
        if self.alt_method:
            M = sym.Matrix([self.ke]).jacobian(self.q.diff(t)).diff(t).jacobian(self.q.diff(t,2))
        else:
            if isinstance(self.frame,HomogenousFrame):
                T = self.frame if sum(self.com_pos)==0 else self.frame.Translate(*self.com_pos)
                Js = T.ManipJacobian(self.q)
                Js = self._trigsimp(Js) if self.simplify else Js
                IA = self._trigsimp(T.InvAdjoint()) if self.simplify else T.InvAdjoint()
                Jb = self._trigsimp(IA*Js) if self.simplify else IA*Js
                M = Jb.T*self.M_e*Jb
            if isinstance(self.frame,ReferenceFrame):
                M_rr = sym.eye(3)*self.M_e[0,0]
                r = Wedge(self.com_pos)
                M_rtheta = -self.M_e[0,0]*self.frame.A*r*self.frame.Gb
                M_thetatheta = self.frame.Gb.T*(self.M_e[3:,3:] + self.M_e[0,0]*r.T*r)*self.frame.Gb
                M = sym.BlockMatrix([[M_rr,M_rtheta],[M_rtheta.T,M_thetatheta]]).as_explicit()   
            M = M[self._idx,self._idx]    
        return self._trigsimp(M) if self.simplify else M


    @property
    @cache
    def pe(self):
        point = self.frame.transform_point(self.com_pos)
        h = -(point.T*self.grav_vec)[0]
        return h*self.M_e[0,0] if h != 0 else 0
    
    def _trigsimp(self,expr):
        return sym.trigsimp(sym.powsimp(sym.cancel(sym.expand(expr))))


from functools import cache
from moyra.frames.base_reference_frame import BaseReferenceFrame
import sympy as sym
from sympy.abc import t
import sympy.physics.mechanics as me
from moyra.helper_funcs import Vee, Vee4, Wedge, Wedge4


class HomogenousFrame(BaseReferenceFrame):

    def __init__(self,E=None):
        E = sym.eye(4) if E is None else E
        super(HomogenousFrame, self).__init__(E[:3,:3],E[:3,3])
    
    @property
    @cache
    def E(self):
        E = sym.eye(4)
        E[:3,:3] = self.A
        E[:3,3] = self.R
        return E


    def BodyJacobian(self,q):
        return self.InvAdjoint()*self.ManipJacobian(q)

    def __mul__(self,other):
        if isinstance(other,HomogenousFrame):
            return HomogenousFrame(self.E*other.E)  
        elif isinstance(other,sym.MutableDenseMatrix):
            return HomogenousFrame(self.E*other)
        else:
            raise TypeError(f'Can not multiple a Homogenous Transform by type {type(other)}')

    def Inverse(self):
        E = sym.eye(4)
        E[:3,:3] = self.A.T
        E[:3,3] = -self.A.T*self.R
        return HomogenousFrame(E)

    def ManipJacobian(self,q):
        inv = self.Inverse().E
        J = sym.zeros(6,len(q))
        for i,qi in enumerate(q):
            J[:,i] = Vee4(self.E.diff(qi)*inv)
        return J

    def Adjoint(self):
        E = sym.zeros(6,6)
        E[:3,:3] = self.A
        E[3:,3:] = self.A
        E[:3,3:] = Wedge(self.R)*self.A.T
        return E

    def InvAdjoint(self):
        E = sym.zeros(6,6)
        E[:3,:3] = self.A.T
        E[3:,3:] = self.A.T
        E[:3,3:] = -self.A.T*Wedge(self.R)
        return E

    def Diff(self):
        E = sym.eye(4)
        E[:3,:3] = self.A.diff(t)
        E[:3,3] = self.R.diff(t)
        return E

    def BodyVelocity(self):
        V = sym.ones(6,1)
        V[:3,0] = self.A.T*self.R.diff(t)

        # Angular velocities skew symetric matrix
        S = self.A.T*self.A.diff(t)
        V[3,0] = S[2,1]
        V[4,0] = S[0,2]
        V[5,0] = S[1,0]
        return V

    def SpatialVelocity(self):
        return self.Diff()*self.Inverse()

    def PuesdoSpatialFrame(self):
        E = self.E.copy()
        E[:3,:3] = sym.eye(3)
        return HomogenousFrame(self.q, E)
    
    def Rotate(self,A):
        H = sym.eye(4)
        H[:3,:3]=A
        return HomogenousFrame(self.E*H)
    
    def R_x_small(self,angle,ExpandCosine=True):
        H = sym.eye(4)
        H[1,1] = H[2,2] = 1-sym.Rational(1,2)*angle**2 if ExpandCosine else 1
        H[1,2] = -angle
        H[2,1] = angle
        return HomogenousFrame(self.E*H)

    def R_x(self,angle):
        H = sym.eye(4)
        H[:3,:3]=sym.rot_axis1(-angle)
        return HomogenousFrame(self.E*H)
    
    def R_y_small(self,angle,ExpandCosine=True):
        H = sym.eye(4)
        H[0,0] = H[2,2] = 1-sym.Rational(1,2)*angle**2 if ExpandCosine else 1
        H[0,2] = angle
        H[2,0] = -angle
        return HomogenousFrame(self.E*H)

    def R_y(self,angle):
        H = sym.eye(4)
        H[:3,:3]=sym.rot_axis2(-angle)
        return HomogenousFrame(self.E*H)
    
    def R_z_small(self,angle,ExpandCosine=True):
        H = sym.eye(4)
        H[0,0] = H[1,1] = 1-sym.Rational(1,2)*angle**2 if ExpandCosine else 1
        H[0,1] = -angle
        H[1,0] = angle
        return HomogenousFrame(self.E*H)

    def R_z(self,angle):
        H = sym.eye(4)
        H[:3,:3]=sym.rot_axis3(-angle)
        return HomogenousFrame(self.E*H)

    def R_rodriguez(self,vector,angle):
        H = sym.eye(4)
        H[:3,:3] += sym.sin(angle)*Wedge(vector) 
        H[:3,:3] += (1-sym.cos(angle))*Wedge(vector)**2
        return HomogenousFrame(self.E*H)

    def R_rodriguez_params(self,qs):
        H = sym.eye(4)
        qs = sym.Matrix(qs)
        R = sym.simplify(sym.eye(3)+2/(1+qs.dot(qs))*(Wedge(qs)+Wedge(qs)*Wedge(qs)))
        H[:3,:3] = R
        return HomogenousFrame(self.E*H)

    def R_euler_params(self,qs):
        H = sym.eye(4)
        qs = sym.Matrix(qs)
        E = sym.Matrix([[-qs[1],qs[0],-qs[3],qs[2]],[-qs[2],qs[3],qs[0],-qs[1]],[-qs[3],-qs[2],qs[1],qs[0]]])
        G = sym.Matrix([[-qs[1],qs[0],qs[3],-qs[2]],[-qs[2],-qs[3],qs[0],qs[1]],[-qs[3],qs[2],-qs[1],qs[0]]])
        H[:3,:3] = E*G.T
        return HomogenousFrame(self.E*H)

    def Translate(self,x,y,z):
        H = sym.eye(4)
        H[:3,3] = sym.Matrix([x,y,z])
        return HomogenousFrame(self.E*H)

    def simplify(self):
        return HomogenousFrame(sym.simplify(self.E))

    def diff(self,*args):  
        return HomogenousFrame(self.E.diff(*args))

    def subs(self,*args):  
        return HomogenousFrame(self.E.subs(*args))

    def msubs(self,*args):  
        return HomogenousFrame(me.msubs(self.E,*args))
